import static org.quartz.CronScheduleBuilder.cronSchedule;
import static org.quartz.JobBuilder.newJob;
import static org.quartz.TriggerBuilder.newTrigger;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.Iterator;
import java.util.TimeZone;
import java.util.TreeMap;

import org.quartz.CronTrigger;
import org.quartz.Job;
import org.quartz.JobDetail;
import org.quartz.Scheduler;
import org.quartz.SchedulerFactory;
import org.quartz.impl.StdSchedulerFactory;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Client that constructs and runs a schedule of data collection jobs based on a configuration file.
 * 
 * @author fhogenboom
 */
public class Client
{

	private String                  cron_ptrn, date_stop, date_strt, form_date, time_zone;
	private TreeMap<Object, Object> comp;

	private static Logger           _log = LoggerFactory.getLogger(Client.class);

	/**
	 * Main method that constructs and runs a {@link Client}.
	 * 
	 * @param args Run-time arguments.
	 */
	public static void main(String[] args)
	{
		showLogo();
		if (args.length > 0)
		{
			Util.setConfig(args[0]);
			Client client = new Client();
			client.run();
		}
		else
		{
			_log.error("Unable to locate configuration file");
		}
	}
	
	/**
	 * Displays an ASCII-art styled logo.
	 */
	private static void showLogo()
	{
		System.out.println("                                                                                                                       ");
		System.out.println("                                                                      `:+yhmNNMNNmhs:                                  ");
		System.out.println("                                                                   -odMMMMMNmdddmNMMNo                                 ");
		System.out.println("                                                                `/dMMMmy+-`       --                                   ");
		System.out.println("                                                               -dMMNy:   ``-////-`                                     ");
		System.out.println("                                                              /NMMh-  `:sdNMMMMMMNh/`                                  ");
		System.out.println("                                                              -+yo  -smMMNmy+//+hNMMd.                                 ");
		System.out.println("                                                                 `:hMMMmo-       -mMMm`                                ");
		System.out.println("                                                                -hMMNh:`          -NMMs                                ");
		System.out.println("         `./ossso/-`+sss .sss-`.:+oss     `-/ossso/-`+sss     `sNMMh:              hMMm       ./ossssssssssssssssso    ");
		System.out.println("       .odNMMMMMMMMmNMMM :MMMddNMMMMN   .smMMMMMMMMMmNMMM    .dMMN+`               sMMN`    `sNMMMMMMMMMMMMMMMMMMMm    ");
		System.out.println("      +mMMNh+:-:/ymMMMMM :MMMMMmy+/:-  +NMMmy/:-:/ymMMMMM   -mMMd-./oo+-`          yMMN     oMMMo:-----------------    ");
		System.out.println("     /MMMhhdds`   .sMMMM :MMMMo.      +MMMy.       .sMMMM  `mMMm-sNMMMMMd:        `NMMy     yMMM:````````              ");
		System.out.println("     dMMN-NMMN.     dMMM :MMMs        mMMm`          mMMM  +MMM:sMMMMMMMMN.       oMMM:     .dMMNmmmmmmmdhs:           ");
		System.out.println("     dMMN`-//.      mMMM :MMM+        mMMm`    .shs.`mMMM  yMMN dMMMMMMMMM:      /MMMs       `/yhdddddddmMMMy          ");
		System.out.println("     /MMMh-       .yMMMM :MMM+        /MMMh-   :mNmohMMMM  +MMM//NMMMMMMMh`    `+NMMy`           ````````sMMM-         ");
		System.out.println("      /mMMNh+/::+yNMMMMM :MMM+         /mMMNho/:+sdNMMMMM  `hMMNo/ymNMNd+`    :dMMNo``::::::::::::::::::/dMMN.         ");
		System.out.println("       .+dNMMMMMMMMdNMMM :MMM+          .+dNMMMMMMMNdNMMM   `oNMMNhs+/-.`.-:odMMMh-  .MMMMMMMMMMMMMMMMMMMMMd:          ");
		System.out.println("         `.:+ooo+:.`/ooo .ooo-            `.:+ooo+:.`hMMM     `+hNMMMNNNNNNMMMms-    `ooooooooooooooooooo/.`           ");
		System.out.println("                                                     dMMN        `-+oyhhhhys/-`                                        ");
		System.out.println("                                                    :MMMy                                                              ");
		System.out.println("                                                  .oNMMm.                                                              ");
		System.out.println("                                               /mNMMMMs`                                                               ");
		System.out.println("                                               +NNdy/`                                                                 ");
		System.out.println("                                                                                                                       ");
		System.out.println("                                        -=* all eyes on your financial data *=-                                        ");
		System.out.println("                                                                                                                       ");
	}

	/**
	 * Constructor.
	 */
	public Client()
	{
		super();
	}

	/**
	 * Configures scheduler by parsing a configuration file.
	 * 
	 * @return {@link Boolean} value indicating whether parsed configuration is valid.
	 */
	private boolean configScheduler()
	{
		_log.info("Configuring scheduler");
		try
		{
			// Load schedule configuration
			cron_ptrn = Util.getString(Util.parseXMLelement("schedule", "cron_ptrn"));
			date_stop = Util.getString(Util.parseXMLelement("schedule", "date_stop"));
			date_strt = Util.getString(Util.parseXMLelement("schedule", "date_strt"));
			form_date = Util.getString(Util.parseXMLelement("schedule", "form_date"));
			time_zone = Util.getString(Util.parseXMLelement("schedule", "time_zone"));

			// Load component configuration
			comp = (TreeMap<Object, Object>) Util.parseXMLparameter("eye", "id", "class");
		}
		catch (Exception e)
		{
			_log.error(e.toString());
		}
		// Return whether parsed configuration is valid
		return (cron_ptrn != null && date_stop != null && date_strt != null && form_date != null
		        && time_zone != null && comp != null);
	}

	/**
	 * Runs scheduler.
	 */
	@SuppressWarnings("unchecked")
	public void run()
	{
		try
		{
			// If parsed configuration is not valid
			if (!configScheduler())
			{
				throw new Exception("Configuration error");
			}
			// Else
			else
			{
				// Initialize scheduler
				_log.info("Initializing scheduler");
				SchedulerFactory sf = new StdSchedulerFactory();
				Scheduler sched = sf.getScheduler();
				// Wait until scheduled start date and time
				try
				{
					_log.info("Waiting");
					DateFormat sdf = new SimpleDateFormat(form_date);
					sdf.setTimeZone(TimeZone.getTimeZone(time_zone));
					Date d_stop = sdf.parse(date_strt);
					Calendar d_strt = Calendar.getInstance();
					d_strt.setTimeZone(TimeZone.getTimeZone(time_zone));
					Thread.sleep(d_stop.getTime() - d_strt.getTime().getTime());
				}
				catch (Exception e)
				{
					_log.error(e.toString());
				}
				// Schedule jobs
				try
				{
					_log.info("Scheduling jobs");
					Iterator<Object> it = comp.keySet().iterator();
					// For each component
					while (it.hasNext())
					{
						try
						{
							// Get id and class
							String id = (String) it.next();
							String cl = (String) comp.get(id);
							_log.info("Adding " + cl + " (id: " + id + ")");
							// Create job
							JobDetail job = newJob((Class<? extends Job>) Class.forName(cl))
							        .withIdentity(id, "group").usingJobData("configured", false)
							        .build();
							// Create trigger
							CronTrigger trigger = newTrigger()
							        .withIdentity("trigger_" + id, "group")
							        .withSchedule(
							                cronSchedule(cron_ptrn).inTimeZone(
							                        TimeZone.getTimeZone(time_zone))).build();
							// Add job and trigger to schedule
							sched.scheduleJob(job, trigger);
						}
						catch (Exception e)
						{
							_log.error(e.toString());
						}
					}
				}
				catch (Exception e)
				{
					_log.error(e.toString());
				}
				// Start scheduler
				_log.info("Starting scheduler");
				sched.start();
				// Wait until scheduled stop date and time
				try
				{
					DateFormat sdf = new SimpleDateFormat(form_date);
					sdf.setTimeZone(TimeZone.getTimeZone(time_zone));
					Date d_stop = sdf.parse(date_stop);
					Calendar d_strt = Calendar.getInstance();
					d_strt.setTimeZone(TimeZone.getTimeZone(time_zone));
					Thread.sleep(d_stop.getTime() - d_strt.getTime().getTime());
				}
				catch (Exception e)
				{
					_log.error(e.toString());
				}
				// Stop scheduler
				_log.info("Stopping scheduler");
				sched.shutdown(true);
			}
		}
		catch (Exception e)
		{
			_log.error(e.toString());
		}
	}

}
